import json
import os
import re
import subprocess
import time
import uuid
from datetime import datetime
from string import punctuation
from typing import Any, Optional, Union

import openai
from emoji import emojize
from pick import pick
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel

import intelliterm
from intelliterm.command_palette import CommandPalette, prompt
from intelliterm.config import config
from intelliterm.console import console
from intelliterm.constants import CODE_THEME, SAVED_CHATS_DIR
from intelliterm.notifications import notification
from intelliterm.prompt import SPECIAL_PROMPTS, Prompt
from intelliterm.types import AutoCopy
from intelliterm.utils import get_file_info, logger, pretty_dict

if "OPENAI_API_KEY" not in os.environ:
    console.error("Missing OPENAI_API_KEY")
    console.print("Add \texport OPENAI_API_KEY = 'API-KEY'\t to your .zshrc or .bashrc")
    quit()

openai.api_key = os.getenv("OPENAI_API_KEY")


class Chat:
    """Class representing a chat client.
    
    Attributes:
        config (Config): Configuration manager.
        _oneshot (bool): If True, exits intelliterm after completion.        
        autocopy (bool): Flag for auto-copying last response to clipboard. 
            Defaults to "off".
        history (list[str]): List of CommandPalette inputs.
        is_completing (bool): Flag indicating whether chat is completing prompt.
        chat_id (str): Unique identifier for the chat session.
        _context (list[ChatPrompt]): Current chat context.
       
    Methods:
        configure(options: list[str]) -> None:
            Manage intelliterm configuration.
        oneshot(toggle: bool) -> None:
            Toggle oneshot client.
        history(input: str) -> None:
            Add user input to history.
        context(prompt: Union[Prompt, list[Prompt]]) -> None:
            Add a single prompt or list of prompts to context.
        last_prompt() -> Optional[ChatPrompt]:
            Return last prompt in context.
        new() -> None:
            Start a new chat (clear context). 
        total_tokens() -> int:
            Return total number of tokens in current chat's context.
        info() -> str:
            Return current chat's info as formatted string.
        file(path: str, prompt: ChatPrompt) -> None:
            Handle file input.
        create_title() -> str:
            Create a title for the current chat's context.
        save() -> None:
            Save current chat.      
        load() -> None:
            Load saved chat.          
        ask(prompt: ChatPrompt, show_input: bool = True) -> None:
            Call model completion on a prompt.
        listen() -> None:
            Listen for new prompts/commands.   
    """
    def __init__(
        self,
        oneshot: bool = False,
        autocopy: AutoCopy = "off",
    ):
        self._oneshot: bool = oneshot
        self.autocopy: AutoCopy = autocopy
        self._history: list[str] = []
        self.is_completing: bool = False
        self.chat_id: str = str(uuid.uuid4())
        self._context: list[Prompt] = [
            Prompt(content=SPECIAL_PROMPTS['SYSTEM'], role="system")
        ]

    def history(self, input: str) -> None:
        """Add user input to history.
        
        Args:
            input (str)
        """
        self._history.append(input)

    def context(self, prompt: Union[Prompt, list[Prompt]]) -> None:
        """Add a single prompt or list of prompts to context.
        
        Args:
            prompt (Union[Prompt, list[Prompt]])
        """
        if isinstance(prompt, list):
            self._context.extend(prompt)
        else:
            self._context.append(prompt)

    def serialize(self) -> dict[str, Any]:
        context = [prompt.serialize() for prompt in self._context]
        obj = {
            "chat_id": self.chat_id,
            "timestamp": str(datetime.now()),
            "_context": context
        }
        return obj

    @classmethod
    def deserialize(cls, json_str: str) -> 'Chat':
        dict = json.loads(json_str)

        dict.pop('timestamp', None)
        chat = Chat()
        chat.__dict__.update(dict)

        new_context: list[Prompt] = []

        for p in dict['_context']:
            prompt = Prompt()
            prompt.__dict__.update(p)
            new_context.append(prompt)

        chat._context = new_context
        return chat

    def configure(self, options: list[str]) -> None:
        """Handle `!config` command.
        
        Args:
            options (list[str]): Configuration options.
        """
        if len(options) > 0:
            match options[0]:
                case "reset":
                    config.reset()
                case "edit":
                    # > !config edit
                    config.edit()
                case _:
                    # > !use <config_name>
                    config_name = options[0]
                    config.use(config_name)
        else:
            # > !cfg
            config.show()

    def oneshot(self, toggle: bool) -> None:
        """Toggle oneshot client.

        Args:
            toggle (bool)
        """
        self._oneshot = toggle

    def last_prompt(self) -> Optional[Prompt]:
        """Return last prompt in context.
        
        Returns:
            Optional[ChatPrompt]: Last prompt in chat. Defaults to None.
        """
        return self._context[-1] if len(self._context) > 1 else None

    def new(self) -> None:
        """Start a new chat (clear context).
        """
        self._context = self._context[:1]     # keep system prompt
        self.chat_id = str(uuid.uuid4())
        console.with_divider(":speech_balloon: new chat")

    def total_tokens(self) -> int:
        """Return total number of tokens in current chat's context.
        
        Returns:
            int: Total number of tokens in chat's context.
        """
        return sum(prompt.token_count() for prompt in self._context)

    def info(self) -> str:
        """Return current chat's info as formatted string.
        
        Returns:
            str: Chat info.
        """
        info = ""
        last_prompt = self.last_prompt()
        total_tokens = self.total_tokens()

        if last_prompt:
            info += f"[bold][{config.get('accent_color')}]:gear: {config.active().name} "
            info += f"[reset]([bold]{last_prompt.token_count()} "
            info += f"[reset]token{'s' if last_prompt.token_count() > 1 else ''}, "
            info += f"[bold]{total_tokens} [reset]total)"
        return info

    def file(self, path: str, prompt: Prompt) -> None:
        """Handle file input.
        
        Args:
            path (str): Path to file.
            prompt (ChatPrompt): (ie: what to do with file)
        """
        path = path.replace("~", os.environ['HOME'])

        if os.environ['HOME'] not in path:
            # use cwd
            path = os.path.join(os.getcwd(), path)

        if os.path.exists(path):
            if os.path.isfile(path):
                with open(path, "r") as file:
                    file_info = get_file_info(path)
                    console.print(
                        Panel(
                            pretty_dict(file_info),
                            title="[bold]:open_file_folder: file",
                        )
                    )
                    prompt.content += file.read()
                    self.ask(prompt, show_input=False)
            else:
                # is a dir / not a file
                return
        else:
            console.error(f"{path} does not exist")

    def create_title(self) -> str:
        """Create a title for the current chat's context.
        
        Returns:
            str: Generated title.
        """
        response = openai.ChatCompletion.create(
            model=config.get("model"),
            messages=([prompt.to_message() for prompt in self._context[1:]] +
                      [Prompt(content=SPECIAL_PROMPTS['CHAT_TITLE']).to_message()]),
            temperature=float(config.get("temperature")),
            presence_penalty=float(config.get("presence_penalty")),
            frequency_penalty=float(config.get("frequency_penalty"))
        )
        title: str = str(response.choices[0].message.content).strip(punctuation)
        title = re.sub(r'[\\/*?:"<>|]', "", title)

        # Duplicate check
        existing_titles = set()
        for filename in os.listdir(SAVED_CHATS_DIR):
            if filename.endswith(".json"):
                existing_titles.add(filename[:-5])

        if title in existing_titles:
            highest_num = 0
            for existing_title in existing_titles:
                match = re.search(rf"{title}_(\d+)$", existing_title)
                if match:
                    number = int(match.group(1))
                    highest_num = max(highest_num, number)

            title = f"{title}_{highest_num + 1}"

        return title

    def save(self) -> None:
        """Save current chat.
        """
        def check_if_saved() -> tuple[bool, Optional[str]]:
            for file_name in os.listdir(SAVED_CHATS_DIR):
                if file_name.endswith(".json"):
                    file_path = os.path.join(SAVED_CHATS_DIR, file_name)
                    with open(file_path) as file:
                        chat = json.loads(file.read())
                        if chat['chat_id'] == self.chat_id:
                            return (True, file_name)
            return (False, None)

        chat_empty = len(self._context) == 1

        if chat_empty:
            notification.emit("Nothing to save (chat empty)")
            return

        saved, saved_name = check_if_saved()

        if saved and saved_name:
            # exists, update existing chat
            file_path = os.path.join(SAVED_CHATS_DIR, saved_name)
        else:
            # does not exist, create new chat
            title = self.create_title()
            file_name = f"{title}.json"
            file_path = os.path.join(SAVED_CHATS_DIR, file_name)

        with open(file_path, "w+") as file:
            chat_dict = self.serialize()
            json.dump(chat_dict, file, indent=4)

            file_path = file_path.replace(os.environ['HOME'], '~')
            logger.info(f"Saved chat ${self.chat_id}: ${file_path}")
            notification.emit(f"Saved chat: ${file_path}")

    # TODO(add test)
    def load(self) -> None:
        """Load a saved chat.
        """
        file_names: list[str] = []
        chats: list[str] = []

        if os.path.exists(SAVED_CHATS_DIR):
            for file_name in os.listdir(SAVED_CHATS_DIR):
                file_path = os.path.join(SAVED_CHATS_DIR, file_name)

                if os.path.isfile(file_path) and file_name.endswith(".json"):
                    with open(file_path, "r") as file:
                        file_names.append(file_name.split(".")[0])
                        chats.append(file.read())
            if file_names:
                file_name, i = pick(
                    file_names,
                    title="Load chat: ",
                    indicator=emojize(":speech_balloon:"),
                )
                if i is not None:
                    selected_chat = Chat.deserialize(chats[i])
                    self.__dict__.update(selected_chat.__dict__)
                    # self.chat_id = selected_chat.chat_id
                    # self._context = selected_chat._context
                    notification.emit(f'Loaded "{file_name}"')
            else:
                notification.emit("No saved chats")
        else:
            notification.emit("No saved chats")

    def ask(self, prompt: Prompt, show_input: bool = True) -> None:
        """Call model completion on a prompt.
        
        Args:
            prompt (ChatPrompt)
            show_input (bool): Show/hide input before completion. Defaults to True.
        """
        prompt_message = prompt.to_message()

        console.clear()
        logger.info(prompt_message)

        self.context(prompt)

        if not self._oneshot and show_input:
            console.print(
                Panel(
                    Markdown(prompt.content, code_theme=CODE_THEME),
                    title=self.info(),
                    title_align="right",
                )
            )

        markdown = Markdown("", code_theme=CODE_THEME)
        full_content = ""

        try:
            self.is_completing = True
            start_time = time.time()
            response = openai.ChatCompletion.create(
                model=config.get("model"),
                messages=[prompt.to_message() for prompt in self._context],
                temperature=float(config.get("temperature")),
                presence_penalty=float(config.get("presence_penalty")),
                frequency_penalty=float(config.get("frequency_penalty")),
                stream=True,
            )

            messages = []

            with Live(
                markdown,
                console=console,
                transient=True,
                refresh_per_second=40,
                vertical_overflow="visible"
            ) as live:
                for chunk in response:
                    message = chunk['choices'][0]['delta']
                    messages.append(message)

                    full_content = "".join([m.get("content", "") for m in messages])
                    markdown = Markdown(
                        full_content,
                        code_theme=CODE_THEME,
                    )
                    live.update(markdown)
            response.close()
        except openai.InvalidRequestError as e:
            console.print("openai:", e)
            return None
        except Exception as e:
            console.print(e)
        except (KeyboardInterrupt, EOFError):
            if self.is_completing:
                console.with_divider(":stop_button: aborted")
                return None
        finally:
            self.is_completing = False
            self.context(
                Prompt(
                    content=full_content,
                    role="assistant",
                    took=time.time() - start_time
                )
            )
            console.print(markdown)

        last_prompt = self.last_prompt()

        if last_prompt:
            match self.autocopy:
                case "all":
                    last_prompt.copy()
                case "code":
                    last_prompt.copy(["code"])
                case "off":
                    pass
        else:
            console.with_divider()

        if not self._oneshot:
            self.listen()

    def listen(self) -> None:
        """Listen for new prompts/commands.
        """
        while True:
            try:
                input = prompt()

                if input is not None and len(input) > 0:
                    self.history(input)

                    if input.startswith(CommandPalette.TRIGGER):
                        parts = input[1:].split()
                        alias, options = parts[0], parts[1:]

                        command = CommandPalette.get_command(alias)
                        last_prompt = self.last_prompt()

                        if command:
                            match command.name:
                                case "version":
                                    console.print(f"[reset]{intelliterm.__version__}")
                                case "help":
                                    CommandPalette.help()
                                case "config":
                                    self.configure(options)
                                case "info":
                                    console.print(self.info())
                                case "save":
                                    self.save()
                                case "load":
                                    self.load()
                                case "copy":
                                    if last_prompt:
                                        last_prompt.copy(options)
                                    else:
                                        notification.emit("Nothing to copy")
                                case "run":
                                    if last_prompt:
                                        if last_prompt.parse_code() == []:
                                            notification.emit("No code to run")
                                        else:
                                            last_prompt.parse_code()[0].run()
                                    else:
                                        notification.emit("No code to run")
                                case "file":
                                    if len(options) > 0:
                                        self.file(
                                            options[0],
                                            Prompt(content="".join(options[1:]))
                                        )
                                    else:
                                        console.error("No file specified")
                                        console.print(command.hint())
                                case "new":
                                    self.new()
                                case "shell":
                                    if options and len(options) > 0:
                                        try:
                                            subprocess.run(options, shell=True)
                                        except subprocess.SubprocessError as e:
                                            console.print(e)
                                    else:
                                        console.error("No shell command specified")
                                        console.print(command.hint())
                                case "quit":
                                    quit()
                        else:
                            CommandPalette.unrecognized(alias)
                    else:
                        self.ask(Prompt(content=input))
            except (KeyboardInterrupt, EOFError):
                if self.is_completing:
                    pass
                else:
                    # otherwise: quit intelliterm
                    quit()



import os
import re
from typing import Any, Mapping, Optional, cast

from rich.console import Console as C
from rich.style import StyleType
from rich.theme import Theme

from intelliterm.styling import COLORS, DANGER_STYLE, INFO_STYLE, WARNING_STYLE


class Console(C):
    """A class for printing formatted messages to the console.

    Inherits from rich.console's `Console`.

    Methods:
        info(*objects: Any) -> None:
            Prints the given objects to the console as information.
        warning(*objects: Any) -> None: 
            Prints the given objects to the console as a warning.
        error(*objects: Any) -> None: 
            Prints the given objects to the console as an error.
        divider(*objects: Any) -> None:  
            Prints a divider as wide as the terminal window.
    """
    def info(self, *objects: Any, sep: str = " ", end: str = "\n") -> None:
        """Print to console as info.
        """
        message = " ".join(str(o) for o in objects)
        self.print(
            f"[bold blue]i [info][not bold]{message}",
            style=INFO_STYLE,
        )

    def warning(self, *objects: Any, sep: str = " ", end: str = "\n") -> None:
        """Print to console as warning.
        """
        message = " ".join(str(o) for o in objects)
        self.print(
            f":warning: {message}",
            sep=sep,
            end=end,
            style=WARNING_STYLE,
        )

    def error(self, *objects: Any, sep: str = " ", end: str = "\n") -> None:
        """Print to console as error.
        """
        message = " ".join(str(o) for o in objects)
        self.print(
            f":warning: {message}",
            sep=sep,
            end=end,
            style=DANGER_STYLE,
        )

    def with_divider(self, title: Optional[str] = None) -> None:
        """Print a divider as wide as the terminal window.
        
        Args:
            title (Optional[str]): Title to center in divider. Defaults to None.
        """
        if title:
            has_emoji = re.search(r":\w+:", title)
            emoji_offset = len(has_emoji.group()) if has_emoji else 0
            offset = (len(title) - emoji_offset + 2) // 2

            self.print(
                ("-" * (os.get_terminal_size().columns // 2 - offset - 2)),
                f"[bold]{title}",
                ("[reset]" + "-" * (os.get_terminal_size().columns // 2 - offset - 1)),
            )
        else:
            self.print('-' * os.get_terminal_size().columns)


console = Console(theme=Theme(cast(Mapping[str, StyleType] | None, COLORS)))



import os
import re
from typing import Any, Generator, Optional

from prompt_toolkit import HTML, PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import Completer, Completion, FuzzyCompleter
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import Validator
from rich.columns import Columns
from rich.panel import Panel

import intelliterm
from intelliterm.console import console
from intelliterm.notifications import notification


key_bindings = KeyBindings()


# TODO(implement): add useful bindings
@key_bindings.add("escape", "c")     # option + c
def exit_(event: Any) -> None:
    """
        Pressing Ctrl-d will exit the user interface.
        """
    event.app.exit()


# TODO(possibly): persist prompt history across sessions
def get_history() -> None:
    pass


session: PromptSession = PromptSession()     #history=)


def bottom_toolbar() -> Any:
    from intelliterm.config import config

    buffer = session.default_buffer
    input = buffer.text
    curr_index = input.count(" ")
    alias, *args = input.split(" ")

    output = ""

    if not notification.read:
        output = notification.message
    else:
        output = f"<strong>{config.active().name}</strong>"

        if input == CommandPalette.TRIGGER:
            output = "Enter a command"
        if CommandPalette.is_valid_input(input):
            command = CommandPalette.get_command(alias)

            if command:
                output = alias + " "

                if command.args:
                    output += " | ".join(
                        arg.name if arg.is_option else f"&lt;{arg.name}&gt;"
                        for arg in command.args
                    )
            if output.startswith(CommandPalette.TRIGGER):
                output_groups = output.split(" ", 1)

                try:
                    output_groups[curr_index
                                  ] = f"<strong>{output_groups[curr_index]}</strong>"
                    output = " ".join(output_groups)
                except IndexError:
                    pass
        else:
            if input != CommandPalette.TRIGGER:
                output = f"<strong>{input}</strong> : unknown command"
    return HTML(output)


def rprompt() -> str:
    from intelliterm.config import config
    return f"[{config.active().name}]"


def prompt() -> str:
    """Prompt for user input.
    
    Returns:
        str: User input.
    """
    from intelliterm.config import config

    style: Style = Style.from_dict({
        "": "",
        "caret": f"fg:ansi{config.get('accent_color')}",
        "bottom-toolbar": f"fg:ansi{config.get('accent_color')}",
    })

    return session.prompt(
        message=[
            ("class:caret", ">"),
            ("class:input", " "),
        ],
        auto_suggest=AutoSuggestFromHistory(),
        completer=FuzzyCompleter(CommandCompleter()),
        mouse_support=False,
        key_bindings=key_bindings,
        validator=validator,
        validate_while_typing=False,
        bottom_toolbar=bottom_toolbar,
        placeholder=HTML(f"<ansiblack>  ({config.active().name})</ansiblack>"),
     # rprompt=rprompt,
        style=style,
    )


class CommandArgument:
    """Class defining a command argument/option.
    
    Attributes:
        name (str): Argument name.
        pos (Optional[int]): Argument position.
        is_option (bool): Argument is an option for a command. Defaults to False.
            (ex: "edit" in: !config edit)
    
    """

    # TODO(refactor): remove `is_option` (ie: improve logic separation)
    def __init__(
        self,
        name: str,
        pos: Optional[int] = None,
        is_option: bool = False,
    ) -> None:
        self.name = name
        self.pos = pos
        self.is_option = is_option


class CommandExample:
    """Class defining a command usage CommandExample.
    
    Attributes:
        command (str): Command name.
        args (Optional[list[Argument]]): List of command arguments/options.
    """
    def __init__(self, command: str, args: Optional[list[CommandArgument]]) -> None:
        self.command = command
        self.args = args


class CommandUsage:
    """Class defining a usage CommandExample for a `Command`.
    
    Attributes:
        command (str): Command name.
        description (str): Usage description.
        args (list[Argument]): List of arguments. Defaults to empty list.
        examples (list[CommandExample]): List of examples. Defaults to empty list.
        
    Methods:
        hint() -> str:
            Returns a helpful hint explaining how to use a command.
    """
    def __init__(
        self,
        command: str,
        description: str,
        args: list[CommandArgument] = [],
        examples: list[CommandExample] = [],
    ):
        self.command = command
        self.args = args
        self.description = description
        self.examples = examples

    def hint(self) -> str:
        """Returns a helpful hint explaining how to use a command.
        
        Returns:
            str
        """
        hint = f"[command]{CommandPalette.TRIGGER}{self.command}[reset]"

        for arg in self.args:
            hint += (
                f" [option]{arg.name}[reset]"
                if arg.is_option else f" [arg]<{arg.name}>[reset]"
            )
        hint += f"\n\t\t{self.description}\n"

        if self.examples:
            hint += "\t\t[bold italic]CommandExample:[reset] "

            for example in self.examples:
                hint += f"[command]{CommandPalette.TRIGGER + self.command}[reset]"

                if example.args:
                    for arg in example.args:
                        hint += f"[arg] {arg.name}"
        hint += "[reset]\n\n"
        return hint


class Command:
    """Class defining a CommandPalette command.
    
    Attributes:
        name (str): Command name.
        description (str): Command description.
        aliases (list[str]): Command aliases.
        args (Optional[list[Argument]]): Command arguments. Defaults to None.
        usage (Optional[list[Usage]]): Command usage template. Defaults to None.
            
    Methods:
        hint() -> str:
            Return documentation hint (command options, usage and examples).
    """
    def __init__(
        self,
        name: str,
        description: str,
        aliases: list[str],
        args: Optional[list[CommandArgument]] = None,
        usage: Optional[list[CommandUsage]] = None,
    ) -> None:
        self.name = name
        self.description = description
        self.aliases = aliases
        self.args = args
        self.usage = usage

    def hint(self) -> str:
        """Return documentation hint (command options, usage and examples).
        
        Returns: 
            str
        """
        hint = ""

        if self.args:
            args = [a for a in self.args if not a.is_option]
            options = [a for a in self.args if a.is_option]

            if args:
                hint += "  [bold italic]args:[reset]\n"
                hint += "\n".join(f"\t[arg]<{arg.name}>[reset]" for arg in args) + "\n"
            if options:
                hint += "  [bold italic]options:[reset]\n"
                hint += "\n".join(
                    f"\t[option]{option.name}[reset]" for option in options
                ) + "\n"
        if self.usage:
            hint += "  [bold italic]usage:[reset]\n\t"
            hint += "\t".join(u.hint() for u in self.usage)

        return hint


class CommandPalette:
    """Class representing a command manager that handles intelliterm-specific commands.
    
    Methods:
        get_command(alias: str) -> Optional[Command]:
            Get command matching alias.  
        help() -> None:
            Show available commands and how to use them.
        is_command(input: str) -> bool:
            Utility to check if input starts with the command trigger.
        is_valid_input(input: str) -> bool:
            Check if input matches a valid command alias.
        is_secondary_alias(alias: str) -> bool:
            Check if alias is not a primary alias 
            (ie: is shortened variant, like `h` for `help`).
        unrecognized(alias: str) -> None:
            Handle unrecognized command alias.
    """
    TRIGGER: str = "!"
    COMMAND_REGEX = r"^!(\w+)(?:\s{0,1}(.*))?"

    AVAILABLE_COMMANDS: list[Command] = [
        Command(
            name="help",
            description="Show available commands",
            aliases=["h", "help"],
        ),
        Command(
            name="version",
            description="Show version",
            aliases=["v", "version"],
        ),
        Command(
            name="config",
            description=f"Manage {intelliterm.__name__} configurations",
            args=[
                CommandArgument("name"),
                CommandArgument("edit", is_option=True),
                CommandArgument("reset", is_option=True),
            ],
            aliases=["u", "use", "cfg", "switch", "config"],
            usage=[
                CommandUsage(
                    command="config",
                    description="Show active configuration",
                ),
                CommandUsage(
                    command="use",
                    args=[CommandArgument("name")],
                    description=
                    "[reset]Use / switch to another configuration (case-insensitive)",
                    examples=[
                        CommandExample(command="use", args=[CommandArgument("gpt3")])
                    ],
                ),
                CommandUsage(
                    command="config",
                    args=[CommandArgument("edit", is_option=True)],
                    description=(
                        "Edit configurations file " +
                        f"(via {os.environ.get('EDITOR', 'nano')})"
                    )
                ),
                CommandUsage(
                    command="config",
                    args=[CommandArgument("reset", is_option=True)],
                    description="Reset configuration to defaults"
                ),
            ],
        ),
        Command(
            name="info",
            description="Show chat info",
            aliases=["i", "info"],
        ),
        Command(
            name="new",
            description="Start new chat / clear context",
            aliases=["n", "new"],
        ),
        Command(
            name="save",
            description="Save chat",
            aliases=["s", "save"],
        ),
        Command(
            name="load",
            description="Load chat",
            aliases=["l", "load"],
        ),
        Command(
            name="copy",
            description="Copy response",
            aliases=["c", "copy"],
            args=[CommandArgument("code", is_option=True)]
        ),
        Command(
            name="run",
            description="Run code block in response",
            aliases=["r", "run"],
        ),
        Command(
            name="file",
            description="Input a file",
            aliases=["f", "file"],
            args=[
                CommandArgument("path"),
                CommandArgument("prompt"),
            ],
            usage=[
                CommandUsage(
                    command="file",
                    args=[
                        CommandArgument("path"),
                        CommandArgument("prompt"),
                    ],
                    description="Input file with prompt",
                    examples=[
                        CommandExample(
                            command="file",
                            args=[
                                CommandArgument("file.txt"),
                                CommandArgument("explain this file"),
                            ]
                        )
                    ]
                ),
            ],
        ),
        Command(
            name="shell",
            description="Run a shell command",
            aliases=["os", "shell"],
            args=[CommandArgument("command")],
            usage=[
                CommandUsage(
                    command="shell",
                    args=[CommandArgument("command")],
                    description=
                    f"Run basic shell commands within {intelliterm.__name__}",
                    examples=[
                        CommandExample(command="shell", args=[CommandArgument("ls")])
                    ]
                ),
            ],
        ),
        Command(
            name="quit",
            description=f"Quit {intelliterm.__name__}",
            aliases=["q", "quit"],
        )
    ]

    GROUPED_ALIASES = [command.aliases for command in AVAILABLE_COMMANDS]
    PRIMARY_ALIASES = [
        command.name if command.name in command.aliases else command.aliases[0]
        for command in AVAILABLE_COMMANDS
    ]
    ALL_ALIASES = [alias for command in AVAILABLE_COMMANDS for alias in command.aliases]

    @staticmethod
    def get_command(alias: str) -> Optional[Command]:
        """Get command matching alias.
        
        Args: 
            alias (str): Command alias.
            
        Returns:
            Optional(Command): Command if it exists, otherwise None.
        """
        alias = alias.replace(CommandPalette.TRIGGER, "")
        return next(
            (
                command for command in CommandPalette.AVAILABLE_COMMANDS
                if alias in command.aliases
            ),
            None,
        )

    @staticmethod
    def get_aliases(command_name: str) -> list[str]:
        """Get all aliases for command with name.
        
        Args:
            command_name (str): Command name.
            
        Returns:
            list[str]: List of aliases associated with command.
        """
        aliases = []
        command = CommandPalette.get_command(command_name)

        if command:
            aliases = command.aliases
        return aliases

    @staticmethod
    def help() -> None:
        """Show available commands and how to use them.
        """
        help_message = ""

        for i, command in enumerate(CommandPalette.AVAILABLE_COMMANDS):
            aliases = [CommandPalette.TRIGGER + alias for alias in command.aliases]
            help_message += f"[command]{', '.join(aliases)}"
            help_message += f"[reset] : {command.description}"
            help_message += "\n" if i != len(
                CommandPalette.AVAILABLE_COMMANDS
            ) - 1 else ""
            help_message += command.hint()
        console.print(Panel(help_message, title="Commands"))

    @staticmethod
    def is_valid_input(input: str) -> bool:
        """Check if input matches a valid command alias.
        
        Args:
            input (str): User input.
            
        Returns:
            bool
        """
        if input.startswith(CommandPalette.TRIGGER):
            match = re.match(CommandPalette.COMMAND_REGEX, input)     # is command
            return match.group(1) in CommandPalette.ALL_ALIASES if match else False
        else:
            return True     # is regular text

    @staticmethod
    def unrecognized(alias: str) -> None:
        """Handle unrecognized command alias.
        """
        grouped_commands = [
            " ".join(["!" + alias for alias in command.aliases])
            for command in CommandPalette.AVAILABLE_COMMANDS
        ]
        columns = [Panel(group) for group in grouped_commands]
        console.print(Columns(columns))

        error = f"{CommandPalette.TRIGGER + alias}[reset] "
        error += "is not a command.\n"
        error += "Enter [command]!help"
        error += "[reset] or [command]!h[reset] "
        error += "to see available commands."

        console.error(error)

    @staticmethod
    def is_secondary_alias(alias: str) -> bool:
        """Check if alias is not a primary alias 
            (ie: is shortened variant, like `h` for `help`).
           
            Returns:
                bool

        """
        return alias not in CommandPalette.PRIMARY_ALIASES


class CommandCompleter(Completer):
    """Auto-completion for CommandPalette.
    """
    def get_completions(self, document: Any, complete_event: Any) -> Generator:
        from intelliterm.config import config

        # current_position: Point = document.get_menu_position()
        # print(current_position)

        word: str = document.current_line.replace(CommandPalette.TRIGGER, "")
        # word: str = document.get_word_before_cursor(pattern=re.compile(COMMAND_REGEX))
        num_words = len(word.split())

        if num_words == 0:
            # First word (!word)
            for command in CommandPalette.AVAILABLE_COMMANDS:
                for alias in command.aliases:
                    if alias.startswith(word):
                        yield Completion(
                            CommandPalette.TRIGGER + alias,
                            start_position=document.get_start_of_document_position(),
                            style=f"fg:ansi{config.get('accent_color')} bg:black",
                        # selected_style="fg:black bg:red ",
                            display_meta=
                            f"(aliased: {CommandPalette.TRIGGER + command.name})"
                            if CommandPalette.is_secondary_alias(alias) else
                            command.description,
                        )


validator = Validator.from_callable(
    CommandPalette.is_valid_input,
    error_message="Invalid command",
)



import subprocess

from prompt_toolkit.shortcuts import confirm

from intelliterm.console import console
from intelliterm.notifications import notification


class Code:
    def __init__(self, language: str, code: str) -> None:
        self.language: str = language
        self.code: str = code
        self._confirmed: bool = False

    # TODO(implement): Create venv + install deps (if necessary to run code snippet)
    # TODO(implement): handle multiple code blocks
    def run(self) -> None:
        """Execute code.
        """
        # TODO(fix): doesn't work yet
        if not self._confirmed:
            self._confirmed = confirm(f"Run {self.language} code?")

        if self._confirmed:
            if self.language in ["bash", "zsh", "shell"]:
                console.print(f"[black]({self.language})\n")
                subprocess.run(self.code.split())
            elif self.language == "python":
                console.print("[black](python)\n")
                subprocess.run(["python3", "-c", self.code])
            elif self.language in ["javascript", "typescript"]:
                console.print("[black](node)\n")
                subprocess.run(["node", "-e", self.code])
            else:
                notification.emit(f"Running {self.language} code is not supported yet")



import argparse
import logging
import os
import sys
from typing import NoReturn, Optional

from intelliterm import __version__
from intelliterm.chat import Chat
from intelliterm.console import console
from intelliterm.prompt import SPECIAL_PROMPTS, Prompt
from intelliterm.utils import (
    intelliterm,
    is_git_diff,
    logger,
    setup_dirs,
    setup_logging,
)


__author__ = "Yulian Kraynyak"
__copyright__ = "Yulian Kraynyak"
__license__ = "MIT"

_logger = logging.getLogger(__name__)


class ArgParser(argparse.ArgumentParser):
    def error(self, message: str) -> NoReturn:
        console.print(f"{intelliterm.__name__}: {message}\n")

        if "file" in message:
            self.print_file_usage()
        else:
            self.print_usage()
        self.exit(2)

    def print_file_usage(self) -> None:
        console.print(
            f"usage: [green]{self.prog}",
            "[reset]-f/--file",
            "[arg]<path>[reset]",
        )

    @staticmethod
    def validate_file(path: str) -> str:
        if not os.path.isfile(path):
            raise argparse.ArgumentTypeError(f"[danger]{path} [reset]is not a file")
        else:
            return path


def parse_args(args: list[str]) -> argparse.Namespace:
    """Parse command line parameters.
    
    Args:
        args (list[str]): CLI parameters as list of strings (for example  `["--help"]`).
            
    Returns:
        argparse.Namespace: CLI parameters namespace.
    """
    parser = ArgParser()
    parser.add_argument(
        "-v",
        "-V",
        "--version",
        action="version",
        help="show version number",
        version=__version__,
    )
    parser.add_argument(
        dest="prompt",
        help="prompt",
        nargs="*",
        type=str,
        metavar="STR",
    )
    parser.add_argument(     # TODO(unfinished)
        "-f",
        "--file",
        dest="file",
        help="input a file",
        type=ArgParser.validate_file,
    )
    parser.add_argument(
        "-m",
        "--mini",
        "--oneshot",
        dest="oneshot",
        action="store_true",
        default=False,
        help=f"complete prompt without entering {intelliterm.__name__}",
    )
    parser.add_argument(
        "-c",
        "--copy",
        dest="autocopy",
        action="store",
        choices=("off", "all", "code"),
        const="all",
        default="off",
        nargs="?",
        help="autocopy response to clipboard",
    )

    return parser.parse_args(args)


def main(_args: list[str]) -> None:
    setup_dirs()
    setup_logging()

    logger.info(f"Starting {intelliterm.__name__}")

    args = parse_args(_args)
    chat = Chat(oneshot=args.oneshot, autocopy=args.autocopy)

    if sys.stdin.isatty():
        # is NOT stdin
        prompt: Optional[Prompt] = None

        if args.prompt:
            # $ ai <prompt>
            prompt = Prompt(content=" ".join(args.prompt))

        if args.file:
            # $ ai -f/--file <file>
            with open(args.file, "r") as file:
                prompt = Prompt(
                    content=file.read() + prompt.content if prompt else "",
                    is_file=True,
                )
                _logger.error(f"Inputting file {args.file}")

        if prompt is None or len(prompt.content.strip()) == 0:
            # $ ai
            console.clear()
            chat.listen()
        else:
            chat.ask(prompt)
    else:
        # <stdin> | ai
        # Call model to complete prompt from stdin
        if args.prompt:
            prompt = Prompt(content=" ".join(args.prompt) + sys.stdin.read().strip())
            chat.ask(prompt)
        else:
            prompt = Prompt(content=sys.stdin.read().strip())

            if len(prompt.content) > 0:
                if is_git_diff(prompt.content):
                    prompt.content = SPECIAL_PROMPTS['GIT_DIFF'] + prompt.content
                chat.oneshot(True)
                chat.ask(prompt)
            else:
                console.error("Empty input")


def run() -> None:
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



import os

import platformdirs

import intelliterm


CODE_THEME = "lightbulb"     # (Pygments theme)

APP_AUTHOR = "Yulian Kraynyak"
USER_DATA_DIR = platformdirs.user_data_dir(intelliterm.__name__, APP_AUTHOR)
DOCUMENTS_DIR = platformdirs.user_documents_dir()
SAVED_CHATS_DIR = os.path.join(DOCUMENTS_DIR, intelliterm.__name__, "chats")
LOGS_DIR = os.path.join(DOCUMENTS_DIR, intelliterm.__name__, "logs")



import os
import random
import subprocess
from configparser import ConfigParser, SectionProxy
from typing import Any

from prompt_toolkit.shortcuts import confirm
from rich.columns import Columns
from rich.panel import Panel

from intelliterm.command_palette import CommandPalette
from intelliterm.console import console
from intelliterm.constants import USER_DATA_DIR
from intelliterm.notifications import notification
from intelliterm.utils import TIPS, logger, longest_line, pretty_dict


class Config:
    """Intelliterm configurations manager.
    
    Methods:
        exists() -> bool:
            Check if configuration file exists.
        default() -> ConfigParser:
            Get default configuration.
        active() -> SectionProxy:
            Get active configuration.
        show() -> None:
            Display active and available configurations.
        get(property: Optional[str] = None) -> str:    
            Get property in active configuration.
        edit() -> None:
            Edit configurations file.
        load() -> None:
            Load configurations file. 
        use(config_name: str) -> None:
            Use a different configuration.
        reset() -> None:
            Reset to default configurations file.
    """

    CONFIG_PATH = os.path.join(USER_DATA_DIR, "config.ini")

    def __init__(self) -> None:
        self.config: ConfigParser = ConfigParser()
        self.load()

    @staticmethod
    def exists() -> bool:
        """Check if configuration file exists.
        """
        return os.path.exists(Config.CONFIG_PATH) and os.path.isfile(Config.CONFIG_PATH)

    @staticmethod
    def default() -> ConfigParser:
        """Get default configuration.
        """        
        default_config = ConfigParser(default_section='GPT3')
        default_config['DEFAULT'] = {
            "model": "gpt-3.5-turbo",
            "temperature": "0",
            "presence_penalty": "0",
            "frequency_penalty": "0",
            "accent_color": "blue"
        }
        default_config['GPT3'] = {
            "model": "gpt-3.5-turbo",
        }
        default_config['GPT4'] = {
            "model": "gpt-4",
        }
        default_config['CONFIG'] = {
            "active": 'GPT3',
        }

        return default_config

    def active(self) -> SectionProxy:
        """Get active configuration.
        """
        return self.config[self.config['CONFIG']['active']]

    def show(self) -> None:
        """Display active and available configurations.
        """
        panels: list[Panel] = []

        for section in self.config:
            if section != "CONFIG" and section != "DEFAULT":
                pretty_config = pretty_dict(
                    dict(self.config[section]),
                    colors=section == self.active().name,
                )

                if section == self.active().name:
                    content = (
                        f"[{self.get('accent_color')}][bold]{section}[reset]" + "\n" +
                        ("-" * longest_line(pretty_config) + "\n") + pretty_config
                    )

                    panels.append(
                        Panel(
                            content,
                            border_style=self.get('accent_color'),
                            title="(active)",
                            title_align="right"
                        )
                    )
                else:
                    panels.append(
                        Panel((
                            section + "\n" + ("-" * longest_line(pretty_config)) + "\n"
                        ) + pretty_config)
                    )

        config_command = CommandPalette.get_command('config')

        if config_command:
            if config_command and config_command.args:
                console.print(
                    Panel(
                        Columns(panels),
                        title="[bold]:gear: Configurations",
                        subtitle=(random.choice(TIPS['config'])),
                        title_align="left",
                        subtitle_align="left"
                    )
                )

    def get(self, property: str) -> str:
        """Get property in active configuration.
        
        Args:
            property (str): Property to get from active configuration.
            
        Returns:
            str: Value for property if not None, else, name of active configuration.
        """
        return self.active()[property]

    def edit(self) -> None:
        """Edit configurations file.
        """
        subprocess.run([os.environ.get("EDITOR", "nano"), Config.CONFIG_PATH])
        self.load()
        self.show()

    def validate(self) -> None:
        # TODO(finish)
        required_values: dict[str, Any] = {
            "": {}
        }

        for section, keys in required_values.items():
            if section not in self.config:
                raise Exception(f"Missing section {section} in config")

    def load(self) -> None:
        """Load configurations file.
        """
        if not self.exists():
            os.makedirs(os.path.dirname(Config.CONFIG_PATH), exist_ok=True)
            with open(Config.CONFIG_PATH, "w+") as file:
                self.default().write(file)
                logger.info("Configurations file not found, set to default")
        self.config.read(Config.CONFIG_PATH)

    def use(self, config_name: str) -> None:
        """Use a different configuration.
        
        Args:
            config_name (str): Name of configuration to set as active.
        """
        config_name = config_name.upper()     # case-insensitive

        self.config.read(Config.CONFIG_PATH)

        if config_name != "CONFIG" and self.config.has_section(config_name):
            self.config.set("CONFIG", "active", config_name)

            with open(Config.CONFIG_PATH, "w+") as config_file:
                self.config.write(config_file)
                self.load()
                notification.emit(f"Switched to <strong>{self.active().name}</strong>")
                logger.info(f"{self.active().name} {dict( self.active())}")
        else:
            console.error(f"No configuration named {config_name}")
            logger.info(f"No configuration named {config_name}")
            self.show()

    def reset(self) -> None:
        """Reset to default configurations file.
        """
        yes: bool = confirm("Reset to default configuration?")

        if yes:
            with open(Config.CONFIG_PATH, "w+") as config_file:
                self.default().write(config_file)
                logger.info(f"{Config.CONFIG_PATH} reset to defaults")
            self.load()
            self.show()

    def to_dict(self) -> dict:
        """Transform config to dictionary.

        Returns:
            dict[str, str]
        """
        return {
            section: dict(self.config[section])
            for section in self.config.sections()
        }


config = Config()



import datetime
import logging
import os
import re

import intelliterm
from intelliterm.constants import (
    LOGS_DIR,
    SAVED_CHATS_DIR,
)


TIPS = {
    "config": [
        "[bold]tip:[not bold] " + tip for tip in [
            "switch to other configs with [command]!use [arg]<name>",
            "add/edit configs with [command]!config [option]edit",
        ]
    ]
}


def setup_logging() -> None:
    """Setup logging.
    """
    log_format = (
        "[%(asctime)s] %(levelname)s:%(name)s.%(module)s:%(funcName)s: " + "%(message)s"
    )
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    filename = os.path.join(LOGS_DIR, f"{intelliterm.__name__}_logs_{timestamp}.log")
    logging.basicConfig(
        level=logging.INFO,
        filename=filename,
        format=log_format,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


logger = logging.getLogger(intelliterm.__name__)


def pretty_dict(dict: dict, colors: bool = True) -> str:
    """Prettify dictionary.
       
        Args:
            dict (dict): Dictionary object.
            colors (bool): Use colors. Defaults to True.
            
        Returns:
            str: Formatted string.
       """
    o = ""
    for i, (k, v) in enumerate(dict.items()):
        is_last = i == len(dict) - 1
        o += f"{k}[reset]: "
        o += "[value]" if colors else ""
        o += v + "[reset]"
        if not is_last:
            o += "\n"
    return o


def longest_line(s: str) -> int:
    """Utility to get length of longest line in string.
    """
    s = re.sub(r"\[.*?]", "", s)     # strip formatting tags
    return max(len(line) for line in s.split("\n"))


def get_file_info(path: str) -> dict[str, str]:
    """Get file information for file at path.
    
    Args:
        path (str): Path to file.
            
    Returns:
        dict[str, str]: Dictionary containing file info.
    """
    return {
        "path": path,
        "type": os.path.splitext(path)[1][1:].lower(),
    }


def is_git_diff(content: str) -> bool:
    """Check if content of string is a git diff.
    """
    return content.startswith("diff --git")


def setup_dirs() -> None:
    os.makedirs(SAVED_CHATS_DIR, exist_ok=True)
    os.makedirs(LOGS_DIR, exist_ok=True)



from time import sleep


class Notification:
    def __init__(self) -> None:
        self._message: str = ""
        self._read: bool = True

    @property
    def message(self) -> str:
        self._read = True
        return self._message

    @message.setter
    def message(self, message: str) -> None:
        self._message = message

    @property
    def read(self) -> bool:
        return self._read

    def emit(self, message: str) -> None:
        self.message = message
        sleep(0.01)     # TODO(refactor): currently a workaround for bottom_toolbar
        self._read = False


notification = Notification()



from rich.style import Style

from intelliterm.types import Colors


ANSI_COLORS = {
    'black': '\033[30m',
    'red': '\033[31m',
    'green': '\033[32m',
    'yellow': '\033[33m',
    'blue': '\033[34m',
    'magenta': '\033[33m',
    'cyan': '\033[36m',
    'white': '\033[37m',
    'reset': '\033[0m'
}

COLORS: Colors = {
    "info": "bright_black",
    "warning": "yellow",
    "danger": "red",
    "surface": "#121722",
    "command": "bold green",
    "arg": "not bold blue",
    "option": "not bold magenta",
    "value": "green"
}

INFO_STYLE = Style(color=f"{COLORS['info']}")
WARNING_STYLE = Style(color=f"{COLORS['danger']}")
DANGER_STYLE = Style(color=f"{COLORS['danger']}", bold=True)



# ruff: noqa: E501

import json
import platform
import re
from typing import Any, Literal, Optional, cast

import pyperclip
import tiktoken

import intelliterm
from intelliterm.code import Code
from intelliterm.config import config
from intelliterm.console import console
from intelliterm.notifications import notification

# TODO(improve): extend `bio` with helpful, general info about self for better responses
bio = {
    "os": platform.system(),
}

SPECIAL_PROMPTS = {
    "SYSTEM":
        f"""
        You are {intelliterm.__name__}, 
        a general knowledge and code assistant running via CLI and specializing in short, straight-to-the-point, intuitive answers.
        
        My bio: {bio}
        
        If I ask anything that involves my bio make sure to make clear
        that all information is general/approximate/non-identifying.
        
        If input is a concept: use analogies to related concepts.
        Edit every response to ensure the following rules are always satisfied:
            - Style the following terms as bold in markdown:
                - names of individuals
                - names of companies and corporations
                - names of company products                
                - stock tickers
            - Code snippets must support syntax highlighting
            - Do not explain code that is short or simple, simply 
              respond with the code itself
        Input:
        """.strip(),
    "GIT_DIFF":
        """
        Generate a commit message, max 50 characters, in conventional format:
        """.strip(),
    "CHAT_TITLE":
        """Summarize this in a maximum of 20 characters"""
}

Role = Literal["system", "assistant", "user"]


class Prompt:
    """Class representing a chat prompt.
    
    Attributes:
        role (Role): Role to assume. Defaults to "user".
        content (str): Prompt content.
        is_file (bool): File input flag.
        
    Methods:
        copy(options: Optional[list[str]] = None) -> None:
            Copy prompt content/code to the clipboard.
        to_message() -> dict[str, str]:  
            Transform to OpenAI Chat Completion message.
        token_count() -> int:
            Count number of tokens in prompt content.
        get_code() -> Optional[Code]:
            Extract code from prompt content.
    """
    def __init__(
        self,
        is_file: bool = False,
        content: str = "",
        role: Role = "user",
        took: Optional[float] = None
    ) -> None:
        self.is_file: bool = is_file
        self.content: str = content
        self.role: Role = role
        self.took: float = 0

    def serialize(self) -> dict[str, Any]:
        return self.__dict__

    @classmethod
    def deserialize(cls, json_str: str) -> 'Prompt':
        dict = json.loads(json_str)
        return cast(Prompt, dict)

    def token_count(self) -> int:
        """Count number of tokens in prompt content.

        Returns:
            int: Number of tokens in prompt content.
        """
        encoding = tiktoken.encoding_for_model(config.get("model"))
        num_tokens = len(encoding.encode(self.content))
        return num_tokens

    def copy(self, options: Optional[list[str]] = None) -> None:
        """Copy prompt content or code to clipboard.

        Args:
            options (Optional[list[str]]): 
                Copy command options. Defaults to None.
        """
        def _copy_to_clipboard(text: str) -> None:
            if len(text) == 0:
                console.print("Nothing to copy")
                return
            pyperclip.copy(text)

        if options is not None and len(options) > 0:
            match options[0]:
                case "code":
                    # > `!c code`
                    # Copy code in content.
                    code_blocks = self.parse_code()

                    if len(code_blocks) == 1:
                        _copy_to_clipboard(code_blocks[0].code)
                        console.with_divider(":clipboard: code copied!")
                    elif len(code_blocks) > 1:
                        # TODO(implement): handle multiple code blocks
                        pass
                    else:
                        notification.emit("No code to copy")

        else:
            # > !c
            # Copy all content.
            if len(self.content) > 0:
                _copy_to_clipboard(self.content)
                console.with_divider(":clipboard: copied!")

    def to_message(self) -> dict[str, Any]:
        """Transform to OpenAI Chat Completion message.

        Returns:
            dict[str, Any]
        """
        return {
            "content": self.content,
            "role": self.role
        }

    def parse_code(self) -> list[Code]:
        """Extract code blocks from prompt content.

        Returns: 
            list[Code] 
        """
        found_code_blocks = re.findall(r"```(\w+)\n([\s\S]*?)\n```", self.content)
        code_blocks: list[Code] = []

        for found in found_code_blocks:
            language = found[0]
            code = found[1]
            code_blocks.append(Code(language, code))
        return code_blocks



from typing import Literal, TypedDict


AutoCopy = Literal['off', 'all', 'code']


class Colors(TypedDict):
    """Color definitions (for type hints).
    """
    info: str
    warning: str
    danger: str
    surface: str
    command: str
    arg: str
    option: str
    value: str
